"""
The handler submodule takes care of all the app-defined handlers.
There are two types of handlers:
    Path-based handlers
        These are called when a user requests a URI that matches the
        handlers base-path and pathinfo-regexp.
        These handlers can process data in three ways,
            plain - The system does no processing other than
                    query-string/form parameters and path_info
            xml   - The system parses the request as XML and returns
                    an ElementTree root element. The system then
                    expects the handler to return a content-type and
                    an ElementTree root element. The return value is
                    serialized using XML and sent with the
                    content-type that was returned.
            json  - The system parses the request as a JSON object,
                    and expects the handler to return an object that
                    will be serialized using JSON.

    Query-string handlers
        These are called when the query-string, or FORM data contains
        a special tag: ACTION:<handlername>=<whatever>
        These can then choose to continue processing the resource
        identified by the URI or they can override the processing.
"""
import re
import logging
import WebFrame as WF
from WebFrame import auth

__registry = { "ANY":{} }
__query_registry = {}

class __HDLR:
    def __init__(self, method, format, path, pattern, handler,
                 read_policy, write_policy):
        self.method = method.upper()
        self.format = format
        self.path = path.rstrip("/")
        self.pattern = pattern
        self.handler = handler
        self.read_policy = read_policy
        self.write_policy = write_policy
        if pattern == None:
            self.pattern_re = None
        else:
            self.pattern_re = re.compile(pattern)

    def run(self, ctx):
        # Check policies
        if auth.check_access(ctx, self.read_policy, self.write_policy):
            # Run different variants
            if self.format == "xml":
                WF._run_xml(self.handler, ctx)
            elif self.format == "json":
                WF._run_json(self.handler, ctx)
            else:
                WF._run_plain(self.handler, ctx)

    def match(self, suff):
        if self.pattern_re == None:
            return len(suff) == 0
        else:
            return self.pattern_re.match(suff) != None

class _QUH:
    def __init__(self, action, handler,
                 read_policy, write_policy):
        self.action = action
        self.handler = handler
        self.read_policy = read_policy
        self.write_policy = write_policy

    def run(self, ctx):
        # Check policies
        if auth.check_access(ctx, self.read_policy, self.write_policy):
            # Run handler
            return self.handler(ctx, self.action)

def __register(method, path, hdlr):
    global __registry
    method = method.upper()
    if not method in __registry:
        __registry[method] = {}
    meth = __registry[method]
            
    if path in meth:
        lst = meth[path]
    else:
        lst = []
        meth[path] = lst
            
    lst.append(hdlr)

def find(method, path, suffix=""):
    """Search for a handler that matches a given request method and
    URI path component, then split out the path and a path suffix
    which will serve as an argument to the handler"""
    global __registry
    if len(path) == 0:
        return None, None
    
    method = method.upper()
    if not method in __registry:
        if not "ANY" in __registry:
            return None, None
        else:
            method = "ANY"
        
    path = path.rstrip("/")
    suffix = suffix.rstrip("/")
    if path in __registry[method]:
        lst = __registry[method][path]
    elif path in __registry["ANY"]:
        lst = __registry["ANY"][path]
    else: 
        # Recurse down one level
        path, presuf = path.rsplit("/", 1)
        return find(method, path, presuf+"/"+suffix)
       
    for h in lst:
        if h.match(suffix):
            return h, suffix

    # Recurse down one level
    path, presuf = path.rsplit("/", 1)
    return find(method, path, presuf+"/"+suffix)

def find_query(params):
    """Check if the params have any action for which there is a handler
    registered. Returns first handler found."""
    if WF.get_opt("debug"):
        logging.debug("Searching for query handler within %s", params)
    pref = WF.get_opt("http.query_handler_prefix")
    for i in params:
        if i.startswith(pref):
            action = i[len(pref):]
            if action in __query_registry:
                if WF.get_opt("debug"):
                    logging.debug("Found query handler for '%s'", action)
                return __query_registry[action]

    # None found
    return None

def register_plain(method, path, pattern, handler, read_policy, write_policy):
    """Register a handler that will be dispatched using the "plain" format"""
    h = __HDLR(method, "plain", path, pattern, handler,
               read_policy, write_policy)
    __register(method, path, h)

def register_xml(method, path, pattern, handler, read_policy, write_policy):
    """Register a handler that will be dispatched using the "xml" format"""
    h = __HDLR(method, "xml", path, pattern, handler,
               read_policy, write_policy)
    __register(method, path, h)

def register_json(method, path, pattern, handler, read_policy, write_policy):
    """Register a handler that will be dispatched using the "json" format"""
    h = __HDLR(method, "json", path, pattern, handler,
               read_policy, write_policy)
    __register(method, path, h)


def register_query(action, handler, read_policy, write_policy):
    """Register a handler that triggers on query parameters beginning with
    ACTION:
    """
    __query_registry[action] = _QUH(action, handler,
               read_policy, write_policy)

